
MARKET_DATA_FIELDS = (
    'DAILY',
    'TICK',
)

CACHE_FIELDS = (
    'STATS',
    'PEERS',
    'NEWS',
    'FINANCIALS',
    'EARNINGS',
    'DIVIDENDS',
    'COMPANY',
    'DAILY',
    'TICK',
)

ALL_FIELDS = (
    'QUOTE',
    'COMPOSITION',
    'STATS',
    'PEERS',
    'NEWS',
    'FINANCIALS',
    'EARNINGS',
    'DIVIDENDS',
    'COMPANY',
    'DAILY',
    'TICK',
)


OTHERS = (
    'POPULARETFS'
)


ETF_URL = 'https://etfdailynews.com/etf/%s/'
POPULAR_ETFS_URL = 'http://etfdb.com/compare/volume/'



import pyEX as p
from datetime import timedelta
from functools import lru_cache
from .utils import last_month, last_close
from .fetch import _fetch as fetch, \
                   fetchStats as backfillStats, \
                   fetchPeers as backfillPeers, \
                   fetchNews as backfillNews, \
                   fetchFinancials as backfillFinancials, \
                   fetchEarnings as backfillEarnings, \
                   fetchDividends as backfillDividends, \
                   fetchCompany as backfillCompany


def whichBackfill(field):
    if field == 'DAILY':
        return backfillDaily
    elif field == 'TICK':
        return backfillMinute
    elif field == 'STATS':
        return backfillStats
    elif field == 'PEERS':
        return backfillPeers
    elif field == 'NEWS':
        return backfillNews
    elif field == 'FINANCIALS':
        return backfillFinancials
    elif field == 'EARNINGS':
        return backfillEarnings
    elif field == 'DIVIDENDS':
        return backfillDividends
    elif field == 'COMPANY':
        return backfillCompany
    else:
        raise NotImplemented


def backfillDaily(distributor, symbols, timeframe='5y', **kwargs):
    if len(symbols) > 0:
        return fetch(distributor, p.chartDF, {'timeframe': '5y'}, symbols)
    return []


@lru_cache(None)
def _getRange(_from):
    dates = []
    while _from < last_close():
        dates.append(_from)
        _from += timedelta(days=1)
    return dates


def backfillMinute(distributor, symbols, _from=last_month(), **kwargs):
    dates = _getRange(_from)
    if len(symbols) > 0:
        # make dates the iterable
        for symbol in symbols:
            yield symbol, p.bulkMinuteBarsDF(symbol, dates)



import pyEX as p
import pandas as pd
import requests
from urllib.error import HTTPError
from .define import POPULAR_ETFS_URL, ETF_URL


def fetch(fields, symbols):
    pass


def refetch(field, symbol):
    if field == 'TICK':
        return p.chartDF(symbol, '1d')
    if field == 'FINANCIALS':
        return p.financialsDF(symbol)
    elif field == 'DAILY':
        return p.chartDF(symbol, '5y')
    elif field == 'COMPANY':
        return p.companyDF(symbol)
    elif field == 'EARNINGS':
        return p.earningsDF(symbol)
    elif field == 'DIVIDENDS':
        return p.dividendsDF(symbol)
    elif field == 'NEWS':
        return p.newsDF(symbol)
    elif field == 'STATS':
        return p.stockStatsDF(symbol)
    elif field == 'COMPOSITION':
        return _fetchComposition(symbol)
    elif field == 'PEERS':
        return p.peersDF(symbol)
    raise NotImplementedError('%s - %s' % (field, symbol))


def whichFetch(field):
    if field == 'DAILY':
        return fetchDaily
    elif field == 'TICK':
        return fetchMinute
    elif field == 'STATS':
        return fetchStats
    elif field == 'PEERS':
        return fetchPeers
    elif field == 'NEWS':
        return fetchNews
    elif field == 'FINANCIALS':
        return fetchFinancials
    elif field == 'EARNINGS':
        return fetchEarnings
    elif field == 'DIVIDENDS':
        return fetchDividends
    elif field == 'COMPANY':
        return fetchCompany
    else:
        raise NotImplementedError('%s' % field)


def _fetch(distributor, foo, foo_kwargs, symbols):
    if len(symbols) > 0:
        for symbol, data in distributor.distribute(foo, foo_kwargs or {}, symbols):
            yield symbol, data


def fetchDaily(distributor, symbols, timeframe='1m', **kwargs):
    if len(symbols) > 0:
        return _fetch(distributor, p.chartDF, {'timeframe': timeframe}, symbols)
    return []


def fetchMinute(distributor, symbols, **kwargs):
    if len(symbols) > 0:
        return _fetch(distributor, p.chartDF, {'timeframe': '1d'}, symbols)
    return []


def fetchStats(distributor, symbols, **kwargs):
    if len(symbols) > 0:
        return _fetch(distributor, p.stockStatsDF, {}, symbols)
    return []


def fetchPeers(distributor, symbols, **kwargs):
    if len(symbols) > 0:
        return _fetch(distributor, p.peersDF, {}, symbols)
    return []


def fetchNews(distributor, symbols, **kwargs):
    if len(symbols) > 0:
        return _fetch(distributor, p.newsDF, {}, symbols)
    return []


def fetchFinancials(distributor, symbols, **kwargs):
    if len(symbols) > 0:
        return _fetch(distributor, p.financialsDF, {}, symbols)
    return []


def fetchEarnings(distributor, symbols, **kwargs):
    if len(symbols) > 0:
        return _fetch(distributor, p.earningsDF, {}, symbols)
    return []


def fetchDividends(distributor, symbols, **kwargs):
    if len(symbols) > 0:
        return _fetch(distributor, p.dividendsDF, {}, symbols)
    return []


def fetchCompany(distributor, symbols, **kwargs):
    if len(symbols) > 0:
        return _fetch(distributor, p.companyDF, {}, symbols)
    return []


def _fetchComposition(symbol):
    try:
        comp = pd.read_html(ETF_URL % symbol, attrs={'id': 'etfs-that-own'})[0]
        comp['% of Total'] = comp['% of Total'].str.rstrip('%').astype(float) / 100.0
        comp.columns = ['Symbol', 'Name', 'Percent']
        comp = comp[['Symbol', 'Percent', 'Name']]
        return comp

    except (IndexError, requests.HTTPError, ValueError, HTTPError):
        return pd.DataFrame()


def fetchComposition(distributor, symbols, **kwargs):
    if len(symbols) > 0:
        for symbol, data in distributor.distribute(_fetchComposition, foo_kwargs or {}, symbols):
            yield symbol, data
    return []


def fetchPopularEtfs(**kwargs):
    return pd.read_html(POPULAR_ETFS_URL)[0]



from datetime import datetime
from .log_utils import log
from .utils import never, last_close, this_week, append
from .validate import SKIP_VALIDATION

EXTRA_CAPACITY = ('TICK',)


def _updateTime(field):
    if field in ('DAILY', 'TICK', 'NEWS'):
        return last_close()
    elif field in ('COMPANY', 'DIVIDENDS', 'EARNINGS', 'FINANCIALS', 'PEERS', 'STATS'):
        return this_week()
    elif field in ('QUOTE',):
        return never()


def _appendIfNecessary(lib, symbol, data_new):
    data_orig = lib.read(symbol).data
    if data_orig.empty:
        lib.write(symbol, data_new, metadata={'timestamp': datetime.now()})
    else:
        lib.write(symbol, append(data_orig, data_new), metadata={'timestamp': datetime.now()})


def _skip(field, symbol=None):
    if symbol is None:
        if (field, '*') in SKIP_VALIDATION:
            return True
        return False
    if (field, symbol) in SKIP_VALIDATION or \
       ('*', symbol) in SKIP_VALIDATION:
        return True
    return False


def _write(lib, symbol, data, dateoverride=None):
    if dateoverride:
        date = dateoverride
    else:
        date = datetime.now()
    lib.write(symbol, data, metadata={'timestamp': date})


def _getLib(db, libname):
    if libname not in db.list_libraries():
        log.critical('Initializing %s' % libname)
        db.initialize_library(libname)
        if libname in EXTRA_CAPACITY:
            db.set_quota(libname, 21474836480)
    return db.get_library(libname)



import pandas as pd


NASDAQ_ROOT = 'https://www.nasdaq.com/symbol/%s/option-chain?dateindex=-1&page=%d'
MONTHS = ['January', 'February', 'March', 'April', 'May', 'June', 'July', 'August', 'September', 'October', 'November', 'December']


def trimMonths(df, field='CDate'):
    ret = df[field].str.contains(MONTHS[0])
    for month in MONTHS[1:]:
        ret = ret | df[field].str.contains(month)
    return ~ret


def fetch(ticker='aapl'):
    dfs = []
    i = 1

    while True:
        print('fetching page %d' % i)
        df = pd.read_html(NASDAQ_ROOT % ('aapl', i))[2]
        df.columns = ['CDate', 'CLast', 'CChg', 'CBid', 'CAsk', 'CVol', 'COpenInt', 'Symbol', 'Strike', 'PDate', 'PLast', 'PChg', 'PBid', 'PAsk', 'PVol', 'POpenInt']
        df = df[df.index > 0]
        df = df[trimMonths(df)]

        # wide to long
        df1 = df[['Symbol', 'Strike', 'CDate', 'CLast', 'CChg', 'CBid', 'CAsk', 'CVol', 'COpenInt']]
        df1.columns = ['Symbol', 'Strike', 'Date', 'Last', 'Chg', 'Bid', 'Ask', 'Vol', 'OpenInt']
        df1['Type'] = 'C'

        df2 = df[['Symbol', 'Strike', 'PDate', 'PLast', 'PChg', 'PBid', 'PAsk', 'PVol', 'POpenInt']]
        df2.columns = ['Symbol', 'Strike', 'Date', 'Last', 'Chg', 'Bid', 'Ask', 'Vol', 'OpenInt']
        df2['Type'] = 'P'

        df = pd.concat([df1, df2])

        if i == 1:
            dfs.append(df)
            i += 1
        elif df.equals(dfs[-1]):
            # duplicate page
            break
        else:
            dfs.append(df)
            i += 1

    df = pd.concat(dfs).reset_index(drop=True)
    return df



import pandas as pd
from functools import lru_cache
from .define import ETF_URL
from .utils import symbols_map


@lru_cache(None)
def composition(key):
    comp = pd.read_html(ETF_URL % key, attrs={'id': 'etfs-that-own'})[0]
    comp['% of Total'] = comp['% of Total'].str.rstrip('%').astype(float) / 100.0
    comp.columns = ['Symbol', 'Name', 'Percent']

    comp['Symbol'].apply(lambda x: symbols_map().get(x, x))
    return comp[['Symbol', 'Percent', 'Name']]


def constituents(key):
    return composition(key)['Symbol'].dropna().values.tolist()


def spy():
    '''spy'''
    return composition('spy')


def spy_constituents():
    '''spy'''
    return constituents('spy')


def sp500():
    '''just use spy'''
    return composition('spy')


def sp500_constituents():
    '''just use spy'''
    return constituents('spy')


def djia():
    '''just use dia'''
    return composition('DIA')


def djia_constituents():
    '''just use dia'''
    return constituents('DIA')


def qqq():
    '''qqq'''
    return composition('qqq')


def qqq_constituents():
    '''qqq'''
    return constituents('qqq')


def nasdaq():
    '''just use qqq'''
    return composition('qqq')


def nasdaq_constituents():
    return constituents('qqq')


def russell1000():
    return composition('IWB')


def russell1000_constituents():
    return constituents('IWB')


def russell2000():
    return composition('IWM')


def russell2000_constituents():
    return constituents('IWM')


def russell3000():
    return composition('IWV')


def russell3000_constituents():
    return constituents('IWV')



import numpy as np
import pandas as pd
from functools import lru_cache


@lru_cache(20)
def whichTransform(field):
    if field == 'DAILY':
        return transformDaily
    elif field == 'TICK':
        return transformMinute
    elif field == 'STATS':
        return transformStats
    elif field == 'PEERS':
        return transformPeers
    elif field == 'NEWS':
        return transformNews
    elif field == 'FINANCIALS':
        return transformFinancials
    elif field == 'EARNINGS':
        return transformEarnings
    elif field == 'DIVIDENDS':
        return transformDividends
    elif field == 'COMPANY':
        return transformCompany
    elif field == 'QUOTE':
        return transformQuote
    elif field == 'COMPOSITION':
        return transformComposition
    else:
        raise NotImplementedError('field: %s' % field)


def _stripDt(df):
    for col in df.select_dtypes(np.datetime64):
        df[col] = df[col].astype(str)


def transformDaily(df, symbol, data):
    if df.empty:
        return {}
    df = df.reset_index()
    _stripDt(df)
    df['ticker'] = symbol
    return df[['date', 'ticker', 'open', 'high', 'low', 'close']][-100:].replace({np.nan: None}).to_dict(orient='records')


def transformMinute(df, symbol, data):
    if df.empty:
        return {}
    df = df.reset_index()
    _stripDt(df)
    df['ticker'] = symbol
    return df[['date', 'ticker', 'open', 'high', 'low', 'close']][-100:].replace({np.nan: None}).to_dict(orient='records')


def transformFinancials(df, symbol, data):
    if df.empty:
        return {}
    df = df.reset_index()
    _stripDt(df)
    return df[-100:].replace({np.nan: None}).to_dict(orient='records')


def transformDividends(df, symbol, data):
    if df.empty:
        return {}
    _stripDt(df)
    return df.replace({np.nan: None}).to_dict(orient='records')


def transformCompany(df, symbol, data):
    if df.empty:
        return {}
    _stripDt(df)
    return df.replace({np.nan: None})[
        ['CEO', 'companyName', 'description', 'sector', 'industry', 'issueType', 'exchange', 'website']].reset_index().replace({np.nan: None}).to_dict(orient='records')[0]


def transformQuote(df, symbol, data):
    if df.empty:
        return {}
    _stripDt(df)
    return df.replace({np.nan: None}).to_dict(orient='records')[0]


def transformEarnings(df, symbol, data):
    if df.empty:
        return {}
    _stripDt(df)
    return df.replace({np.nan: None}).to_dict(orient='records')


def transformNews(df, symbol, data):
    if df.empty:
        return {}
    _stripDt(df)
    df = df.replace({np.nan: None})
    if not df.empty:
        df['headline'] = '<a href="' + df['url'] + '">' + df['headline'] + ' [<strong>' + df['source'] + '</strong>]' + '</a>'
        df['summary'] = '<p>' + df['summary'] + '</p>'
        return df[['headline', 'summary']].to_dict(orient='records')
    else:
        return {}


def transformPeers(df, symbol, data):
    if df is not None and not df.empty:
        df = df.replace({np.nan: None})
        infos = pd.concat([data.read(item, 'COMPANY') for item in df.index.values])
    else:
        infos = pd.DataFrame()
    return infos.replace({np.nan: None}).to_dict(orient='records')


def transformStats(df, symbol, data):
    if df.empty:
        return {}
    _stripDt(df)
    return df.replace({np.nan: None}).to_dict(orient='records')


def transformComposition(df, symbol, data):
    if df.empty:
        return {}
    _stripDt(df)
    return df.replace({np.nan: None}).to_dict(orient='records')



import pyEX as p
import pandas as pd
from datetime import datetime
import arctic
from arctic import Arctic
from pymongo.errors import ServerSelectionTimeoutError
from .utils import never, business_days, last_month, yesterday, today
from .data_utils import _getLib, _appendIfNecessary, _updateTime, _skip
from .distributor import Distributer
from .backfill import whichBackfill
from .fetch import whichFetch, refetch
from .log_utils import log, logit
from .define import CACHE_FIELDS as FIELDS


class OfflineLib(object):
    def has_symbol(self, symbol):
        return False


class OfflineDB(object):
    def list_libraries(self):
        return FIELDS

    def initialize_library(self, name):
        return OfflineLib()

    def get_library(self, name):
        return OfflineLib()


class Data(object):
    def __init__(self, dbname, offline=False):
        self.libraries = {}
        self.distributor = Distributer.default()

        if offline:
            self.db = OfflineDB()
            log.critical('WARNING Running in offline mode')
            return

        self.db = Arctic(dbname)

        # initialize databases
        for field in FIELDS:
            try:
                self.libraries[field] = _getLib(self.db, field)
            except (arctic.exceptions.LibraryNotFoundException, ServerSelectionTimeoutError):
                log.critical('Arctic not available, is mongo offline??')
                raise

    def cache(self, symbols=None, fields=None, delete=False):
        fields = fields or FIELDS
        symbols = symbols or p.symbolsDF().index.values.tolist()

        to_delete, to_fill, to_update = self.initialize(symbols, fields)

        if delete:
            # prune data
            self.delete(to_delete)

        self.backfill(to_fill)
        self.update(to_update)
        self.validate()

    def delete(self, to_delete=None):
        # delete data no longer needed
        for field in to_delete:
            for symbol in to_delete[field]:
                log.critical('Deleting %s from %s' % (symbol, field))
                self.libraries[field].delete(symbol)

    def backfill(self, to_fill):
        # backfill data if necessary
        for field in to_fill:
            log.critical('Backfilling %d items' % len(to_fill[field]))
            lib = self.libraries[field]

            for symbol, data in whichBackfill(field)(self.distributor, to_fill[field]):
                log.critical('Filling %s for %s' % (symbol, field))
                data_orig = lib.read(symbol).data
                _appendIfNecessary(lib, symbol, data_orig, data)

    def update(self, to_update):
        # update data if necessary
        for field in to_update:
            log.critical('Updating %d items' % len(to_update[field]))
            lib = self.libraries[field]

            for symbol, data in whichFetch(field)(self.distributor, to_update[field]):
                log.critical('Updating %s for %s' % (symbol, field))
                data_orig = self.libraries[field].read(symbol).data
                _appendIfNecessary(lib, symbol, data_orig, data)

    def initialize(self, symbols=None, fields=None):
        '''setup db'''
        fields = fields or FIELDS
        symbols = symbols or p.symbolsDF().index.values.tolist()

        to_fill = {}
        to_update = {}
        to_delete = {}

        _empty = pd.DataFrame()

        # initialize database and collect what to update
        for field in FIELDS:
            if field not in to_fill:
                to_fill[field] = []
            if field not in to_update:
                to_update[field] = []
            if field not in to_delete:
                to_delete[field] = []

            library = self.libraries[field]
            all_symbols = library.list_symbols()

            for symbol in symbols:
                symbol = symbol.upper()
                if symbol not in all_symbols:
                    log.critical('Initializing %s for %s' % (symbol, field))
                    to_fill[field].append(symbol)
                    library.write(symbol, _empty, metadata={'timestamp': never()})

                else:
                    metadata = library.read_metadata(symbol.upper()).metadata
                    if not metadata or not metadata.get('timestamp'):
                        to_fill[field].append(symbol)
                    elif metadata.get('timestamp', never()) <= never():
                        to_fill[field].append(symbol)
                    elif metadata.get('timestamp', never()) < _updateTime(field):
                        to_update[field].append(symbol)

            for symbol in set(all_symbols) - set(symbols):
                    to_delete[field].append(symbol)
        return to_delete, to_fill, to_update

    def validate(self, symbols=None, fields=None):
        '''look for missing data'''
        fields = fields or FIELDS
        symbols = symbols or p.symbolsDF().index.values.tolist()
        to_refill = {}
        self.initialize(symbols, fields)

        for field in FIELDS:
            tick_start_date = today()
            daily_start_date = today()
            fail_count = 0
            print_fail = False
            dates = business_days(last_month(), yesterday())
            to_refill[field] = []

            if _skip(field):
                continue

            dbs = self.db.list_libraries()
            if field not in dbs:
                log.critical('VALIDATION FAILED %s' % field)
                continue

            lib = self.db.get_library(field)
            all_symbols = lib.list_symbols()

            for symbol in symbols:
                symbol = symbol.upper()

                # if fail count too high, autofail all for speed
                if fail_count > .2 * len(all_symbols):
                    if not print_fail:
                        log.critical('VALIDATION THRESHOLD REACHED for %s' % field)
                        print_fail = True

                    if _skip(field, symbol):
                        continue

                    to_refill[field].append(symbol)
                    if field == 'DAILY':
                        daily_start_date = dates[0]
                    if field == 'TICK':
                        tick_start_date = dates[0]
                    continue

                if _skip(field, symbol):
                    continue

                if symbol not in all_symbols:
                    to_refill[field].append(symbol)
                    log.critical('VALIDATION FAILED %s for %s' % (symbol, field))
                    fail_count += 1
                    continue

                data = lib.read(symbol).data

                if data.empty:
                    log.critical('VALIDATION FAILED - DATA EMPTY %s for %s' % (symbol, field))
                    to_refill[field].append(symbol)
                    fail_count += 1
                    continue

                elif field in ('TICK'):
                    for date in dates:
                        if date not in data.index:
                            log.critical('VALIDATION FAILED - DATA MISSING %s for %s : %s' % (symbol, field, date.strftime('%Y%m%d')))
                            to_refill[field].append(symbol)
                            tick_start_date = min(tick_start_date, date) if tick_start_date is not None else date
                            fail_count += 1
                            break

                elif field in ('Daily'):
                    for date in dates:
                        if date not in data.index:
                            log.critical('VALIDATION FAILED - DATA MISSING %s for %s : %s' % (symbol, field, date.strptime('%Y%m%d')))
                            to_refill[field].append(symbol)
                            daily_start_date = min(daily_start_date, date) if daily_start_date is not None else date
                            fail_count += 1
                            break

        # backfill data if necessary
        for field in to_refill:
            lib = self.libraries[field]

            if field == 'TICK':
                log.critical('Backfilling %d items for %s - %s' % (len(to_refill[field]), field, str(tick_start_date)))
            elif field == 'DAILY':
                log.critical('Backfilling %d items for %s - %s' % (len(to_refill[field]), field, str(daily_start_date)))
            else:
                log.critical('Backfilling %d items for %s' % (len(to_refill[field]), field))

            for symbol, data in whichBackfill(field)(self.distributor, to_refill[field], from_=tick_start_date):
                log.critical('Updating %s for %s' % (symbol, field))

                data_orig = lib.read(symbol).data
                _appendIfNecessary(lib, symbol, data_orig, data)

    def read(self, symbol, field, fetch=True, fill=False):
        field = field.upper()
        symbol = symbol.upper()

        if field in ('QUOTE'):
            # dont cache, instantaneous
            return p.quoteDF(symbol)
        elif field in ('COMPOSITION'):
            return refetch(field, symbol)

        if field not in self.libraries and not fetch:
            return pd.DataFrame()

        l = _getLib(self.db, field)

        if not l.has_symbol(symbol):
            if not fetch:
                return pd.DataFrame()
            df = pd.DataFrame()
        else:
            df = l.read(symbol).data
            metadata = l.read_metadata(symbol).metadata

        if fetch:
            if df.empty or not metadata or not metadata.get('timestamp') or \
               metadata.get('timestamp', never()) <= never() or \
               metadata.get('timestamp', never()) < _updateTime(field):

                df = refetch(field, symbol)
                if fill:
                    l.write(symbol, df, metadata={'timestamp': datetime.now()})
        return df



import pandas as pd
import pyEX as p
import string
from trading_calendars import get_calendar
from datetime import datetime, timedelta, date
from dateutil.relativedelta import relativedelta
from functools import lru_cache


_TRANSLATOR = str.maketrans('', '', string.punctuation)
_OVERRIDES = {
    'PCLN': 'BKNG'
}


@lru_cache(1)
def symbols():
    return p.symbolsList()


@lru_cache(1)
def symbols_map():
    ret = {}
    for x in symbols():
        ret[x] = x
        new_x = x.translate(_TRANSLATOR)
        if new_x not in ret:
            ret[new_x] = x
    for k, v in _OVERRIDES.items():
        ret[k] = v
    return ret


@lru_cache(1)
def today():
    '''today starts at 4pm the previous close'''
    today = date.today()
    return datetime(year=today.year, month=today.month, day=today.day)


@lru_cache(1)
def this_week():
    '''start of week'''
    return today() - timedelta(days=datetime.today().isoweekday() % 7)


@lru_cache(1)
def last_close():
    '''last close'''
    today = date.today()
    close = datetime(year=today.year, month=today.month, day=today.day, hour=16)

    if datetime.now().hour < 16:
        close -= timedelta(days=1)
        if close.weekday() == 5:  # saturday
            return close - timedelta(days=1)
        elif close.weekday() == 6:  # sunday
            return close - timedelta(days=2)
        return close
    return close


@lru_cache(1)
def yesterday():
    '''yesterday is anytime before the previous 4pm close'''
    today = date.today()

    if today.weekday() == 0:  # monday
        return datetime(year=today.year, month=today.month, day=today.day) - timedelta(days=3)
    elif today.weekday() == 6:  # sunday
        return datetime(year=today.year, month=today.month, day=today.day) - timedelta(days=2)
    return datetime(year=today.year, month=today.month, day=today.day) - timedelta(days=1)


@lru_cache(1)
def last_month():
    '''last_month is one month before today'''
    today = date.today()
    last_month = datetime(year=today.year, month=today.month, day=today.day) - relativedelta(months=1)

    if last_month.weekday() == 5:
        last_month -= timedelta(days=1)
    elif last_month.weekday() == 6:
        last_month -= timedelta(days=2)
    return last_month


@lru_cache(1)
def six_months():
    '''six_months is six months before today'''
    today = date.today()
    six_months = datetime(year=today.year, month=today.month, day=today.day) - relativedelta(months=6)

    if six_months.weekday() == 5:
        six_months -= timedelta(days=1)
    elif six_months.weekday() == 6:
        six_months -= timedelta(days=2)
    return six_months


@lru_cache(1)
def three_months():
    '''three_months is three months before today'''
    today = date.today()
    six_months = datetime(year=today.year, month=today.month, day=today.day) - relativedelta(months=3)

    if six_months.weekday() == 5:
        six_months -= timedelta(days=1)
    elif six_months.weekday() == 6:
        six_months -= timedelta(days=2)
    return six_months


@lru_cache(1)
def never():
    '''long long time ago'''
    return datetime(year=1, month=1, day=1)


def append(df1, df2):
    merged = pd.concat([df1, df2])
    return merged[~merged.index.duplicated(keep='first')]


@lru_cache(1)
def holidays():
    return get_calendar('NYSE').regular_holidays.holidays().to_pydatetime().tolist()


@lru_cache(None)
def business_days(start, end=last_close()):
    ret = []
    while start < last_close():
        if start not in holidays() and start.weekday() != 6 and start.weekday() != 5:
            ret.append(start)
        start += timedelta(days=1)
    return ret



import pyEX
from functools import lru_cache
from .data import Data
from .define import ALL_FIELDS
from .transform import whichTransform


class Cache(object):
    def __init__(self, host='localhost', offline=False):
        self.data = Data(host, offline=offline)

    @lru_cache(100)
    def _fetch(self, symbol, field):
        return self._fetch_nocache(symbol, field)

    def _fetch_nocache(self, symbol, field):
        field = field.upper()
        symbol = symbol.upper()

        if field != 'ALL' and field not in ALL_FIELDS:
            raise Exception('Unknown field %s' % field)

        if field == 'ALL':
            fields = ALL_FIELDS
        else:
            fields = [field]

        ret = {}
        for field in fields:
            df = self._fetchDF_nocache(symbol, field)
            ret[field] = whichTransform(field)(df, symbol, self.data)
        return ret

    def fetch(self, symbol, field, cache=True):
        if cache:
            return self._fetch(symbol, field)
        return self._fetch_nocache(symbol, field)

    @lru_cache(100)
    def _fetchDF(self, symbol, field):
        return self._fetchDF_nocache(symbol, field)

    def _fetchDF_nocache(self, symbol, field):
        field = field.upper()
        symbol = symbol.upper()

        if field == 'ALL' and field not in ALL_FIELDS:
            raise Exception('Unknown or Invalid field %s' % field)

        return self.data.read(symbol, field)

    def fetchDF(self, symbol, field, cache=True):
        if cache:
            return self._fetchDF(symbol, field)
        return self._fetchDF_nocache(symbol, field)

    @lru_cache(1)
    def tickers(self):
        return pyEX.symbolsDF()



import pandas as pd
import traceback
from functools import partial
from multiprocessing.pool import ThreadPool
from .log_utils import log


def chunks(l, n):
    '''MISH'''
    if len(l) > 0:
        for i in range(0, len(l), n):
            yield l[i:i + n]


class Distributer(object):
    def __init__(self, kind, chunkSize=20):
        self.kind = kind
        self.chunk_size = chunkSize

    def distribute(self, function, function_kwargs, iterable, starmap=False, skip_if_error=True, max_attempts=3):
        if self.kind == 'thread':
            self.pool = ThreadPool(self.chunk_size)
        else:
            raise NotImplemented

        if isinstance(self.pool, ThreadPool):
            for chunk in chunks(iterable, self.chunk_size):
                attempts = 0
                individually = False

                while attempts < max_attempts:
                    try:
                        if not individually:
                            if starmap:
                                ret = self.pool.starmap(function, chunk)
                            else:
                                ret = self.pool.map(partial(function, **function_kwargs), chunk)
                            attempts = max_attempts
                        else:
                            ret = []
                            for item in chunk:
                                try:
                                    if starmap:
                                        val = function(*item)
                                    else:
                                        val = partial(function, **function_kwargs)(item)
                                    ret.append(val)

                                except Exception as e:
                                    log.critical(e)
                                    log.critical(''.join(traceback.format_exception(None, e, e.__traceback__)))
                                    attempts += 1
                                    if attempts >= max_attempts:
                                        if skip_if_error:
                                            ret.append(pd.DataFrame())
                                        else:
                                            raise e

                    except Exception as e:
                        log.error(e)
                        attempts += 1
                        if attempts >= max_attempts:
                            individually = True
                            attempts = 0

                for i, item in enumerate(ret):
                    yield (chunk[i], item)

        else:
            raise NotImplemented

        self.pool.stop()

    def stop(self):
        self.pool.stop()

    @staticmethod
    def default():
        return Distributer('thread')
